import Foundation

enum FileTypeCategory: String, CaseIterable {
    case image
    case code
    case model
    case text
    case video
    case resource

    var text: String {
        switch self {
        case .image: return "Images"
        case .code: return "Code"
        case .model: return "Models"
        case .text: return "Text"
        case .video: return "Videos"
        case .resource: return "Other"
        }
    }
}

extension FileTypeCategory: Hashable {
    
}

extension FileTypeCategory: Identifiable {

    var id: String {
        rawValue
    }
}

enum FileType {
    case image(ImageFileType)
    case code(CodeFileType)
    case model(ModelFileType)
    case text(TextFileType)
    case video(VideoFileType)
    case other(ResourceFileType)

    init(fileExtension: String?) {
        let ext = fileExtension?.lowercased() ?? ""

        if let image = ImageFileType(fileExtension: ext) {
            self = .image(image)
        } else if let code = CodeFileType(fileExtension: ext) {
            self = .code(code)
        } else if let model = ModelFileType(fileExtension: ext) {
            self = .model(model)
        } else if let text = TextFileType(fileExtension: ext) {
            self = .text(text)
        } else if let video = VideoFileType(fileExtension: ext) {
            self = .video(video)
        } else {
            let resource = ResourceFileType(fileExtension: ext)
            self = .other(resource)
        }
    }

    var fileExtension: String {
        switch self {
        case .image(let type): return type.fileExtension
        case .code(let type): return type.fileExtension
        case .model(let type): return type.fileExtension
        case .text(let type): return type.fileExtension
        case .video(let type): return type.fileExtension
        case .other(let type): return type.fileExtension
        }
    }

    var isImage: Bool {
        if case .image = self {
            return true
        }
        return false
    }

    var isVideo: Bool {
        if case .video = self {
            return true
        }
        return false
    }

    var isTextFile: Bool {
        switch self {
        case .code, .text: return true
        default: return false
        }
    }

    var isOtherFile: Bool {
        switch self {
        case .model, .other: return true
        default: return false
        }
    }

    var videoType: VideoFileType? {
        if case .video(let videoType) = self {
            return videoType
        }
        return nil
    }
}
